from lockfile import FileLock


class Lock(FileLock):
    def __init__(self, file_, pid=None, name=None, season=None, episode=None,
                 port=None):
        FileLock.__init__(self, file_)
        self.pid = str(pid)
        self.name = name
        self.season = season
        self.episode = episode
        self.port = port

    def acquire(self, *args, **kwargs):
        FileLock.acquire(self, *args, **kwargs)
        self._write_data()

    def _write_data(self):
        file_ = open('%s.lock' % self.path, 'w')
        season = self.season if self.season is not None else ""
        episode = self.episode if self.episode is not None else ""

        file_.write(','.join((self.pid, self.name, season, episode,
                              self.port)))
        file_.close()

    def _get_file_data(self):
        try:
            file_ = open('%s.lock' % self.path, 'r')
            data = file_.read()
            parts = data.split(",")
        except IOError:
            parts = []

        return parts

    def is_same_file(self, name, season, episode):
        ret = False
        parts = self._get_file_data()
        if len(parts) > 1:
            if season is None:
                season = ''
            if episode is None:
                episode = ''
            ret = parts[1] == name and parts[2] == season and \
                parts[3] == episode
        return ret

    def get_pid(self):
        pid = self._get_file_data()[0]
        return int(pid) if pid != '' else None




#! /usr/bin/env python2
import argparse
import logging
import sys

from babelfish import Language
from blessings import Terminal
from libtorrent import version as libtorrent_version

from touchandgo.helpers import daemonize
from touchandgo.logger import log_set_up
from touchandgo.search import SearchAndStream


log = logging.getLogger('touchandgo.main')


def main():
    parser = argparse.ArgumentParser(
        description="Command line tool to watch torrents")
    parser.add_argument("name", help="The name of the series or movie")

    parser.add_argument("season_number", default=None, nargs="?", type=int,
                        help="Season number")
    parser.add_argument("episode_number", default=None, nargs="?", type=int,
                        help="Episode number")
    parser.add_argument("--sub", nargs='?', default=None,
                        help="Subtitle language")
    parser.add_argument("--serve", action="store_true",
                        help="Do not run VLC")
    parser.add_argument("--quality", nargs='?', default=None,
                        help="quality of the video [normal|hd|fullhd]")
    parser.add_argument("--daemon", action="store_true",
                        help="Daemonize the process"),
    parser.add_argument("--port", "-p", default="8888",
                        help="The port where the stream will be served")
    parser.add_argument("--verbose", action="store_true", default=None,
                        help="Show _all_ the logs")
    parser.add_argument("--player", default='vlc',
                        help="Player to use. vlc|omxplayer|chromecast")
    parser.add_argument("--search", default=None,
                        help="search lib to use (options are 'kat', 'tms', 'strike')")
    parser.add_argument("--nocache", action="store_false", default=True,
                        help="Search for the torrent again"),

    args = parser.parse_args()

    log_set_up(args.verbose)
    log.info("Starting touchandgo")
    log.debug("Running Python %s on %r", sys.version_info, sys.platform)
    log.debug("Libtorrent version: %s", libtorrent_version)

    try:
        if args.sub is not None:
            Language(args.sub)

        touchandgo = SearchAndStream(args.name, season=args.season_number,
                                     episode=args.episode_number,
                                     sub_lang=args.sub, serve=args.serve,
                                     quality=args.quality, port=args.port,
                                     player=args.player, search=args.search,
                                     use_cache=args.nocache)
        if args.daemon:
            def callback():
                touchandgo.serve = True
                touchandgo.watch()
            daemonize(args, callback)
        else:
            term = Terminal()
            with term.fullscreen():
                touchandgo.watch()
    except ValueError as e:
        print(e)

if __name__ == '__main__':
    main()



"""
http range detection based on
 https://github.com/menboku/musicsharer/blob/master/HTTPRangeServer.py
"""
import re
import logging

from SocketServer import ThreadingMixIn
from BaseHTTPServer import HTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler

from json import dumps
from os import fstat
from time import sleep

from touchandgo.constants import STATES
from touchandgo.settings import WAIT_FOR_IT


log = logging.getLogger('touchandgo.stream_server')


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass


class VideoHandler(SimpleHTTPRequestHandler):
    manager = None

    def handle_one_request(self, *args, **kwargs):
        try:
            return SimpleHTTPRequestHandler.handle_one_request(self, *args,
                                                               **kwargs)
        except:
            pass

    def status(self):
        log.debug("returning 200 code")
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        manager = self.manager
        dict_ = {"file": manager.video_file,
                 "elapsed": manager.elapsed_time(),
                 "rates": manager.rates(),
                 "state": STATES[manager.status.state],
                 "peers": manager.status.num_peers,
                 "streaming": manager.streaming}
        data = dumps(dict_)
        self.wfile.write(data)

    def do_GET(self):

        if self.path == "/status":
            self.status()
        else:
            f = self.send_head()
            if f:
                if self.range_from is not None and self.range_to is not None:
                    self.copy_chunk(f, self.wfile)
                else:
                    self.copyfile(f, self.wfile)
                f.close()

    def copy_chunk(self, in_file, out_file):
        def get_piece_length():
            info = self.manager.handle.get_torrent_info()
            length = info.piece_length()
            return length

        def get_blocks_for_range(range_from, range_to):
            length = get_piece_length()
            block_from = self.range_from / length
            block_to = self.range_to / length
            return block_from, block_to

        def is_block_available(block_number):
            status = self.manager.handle.status()
            pieces = status.pieces
            return all(pieces[block_number:block_number+3])

        in_file.seek(self.range_from)
        length = get_piece_length()

        bytes_copied = 0
        blocks = get_blocks_for_range(self.range_from, self.range_to)
        for block_number in range(blocks[0], blocks[1]+1):
            if not is_block_available(block_number):
                log.debug("requesting block %s" % block_number)
                self.manager.strategy.block_requested(block_number)
                sleep(WAIT_FOR_IT)

            while not is_block_available(block_number):
                sleep(WAIT_FOR_IT)
            self.manager.block_served(block_number)
            read_buf = in_file.read(length)
            if len(read_buf) == 0:
                break
            out_file.write(read_buf)
            bytes_copied += len(read_buf)
        return bytes_copied

    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.


        """
        path = self.manager.get_video_path()
        try:
            # Always read in binary mode. Opening files in text mode may
            # cause newline translations, making the actual size of the
            # content transmitted *less* than the content-length!
            f = open(path, 'rb')
        except IOError:
            self.send_error(404, "File not found")
            return None

        fs = fstat(f.fileno())
        total_length = fs[6]
        try:
            self.range_from, self.range_to = parse_range_header(
                self.headers.getheader("Range"), total_length)
        except InvalidRangeHeader:
            # Just serve them the whole file, although it's possibly
            # more correct to return a 4xx error?
            log.warning("Range header parsing failed, "
                            "serving complete file")
            self.range_from = self.range_to = None

        if self.range_from is not None or self.range_to is not None:
            log.debug("returning 206 code")
            self.send_response(206)
        else:
            log.debug("returning 200 code")
            self.send_response(200)
            self.send_header("Accept-Ranges", "bytes")
        guess = self.manager.guess(path)
        self.send_header("Content-Type", guess['mimetype'])
        log.debug("mime type is %s" % guess['mimetype'])
        log.debug("range %s to %s" % (self.range_from, self.range_to))
        if self.range_from is not None or self.range_to is not None:
            # TODO: Should also check that range is within the file size
            self.send_header("Content-Range",
                             "bytes %d-%d/%d" % (self.range_from,
                                                 self.range_to,
                                                 total_length))
            # Add 1 because ranges are inclusive
            self.send_header("Content-Length",
                             (1 + self.range_to - self.range_from))
        else:
            self.send_header("Content-Length", str(total_length))
        self.send_header("Last-Modified",
                         self.date_time_string(fs.st_mtime))
        self.end_headers()
        return f

    def finish(self, *args, **kwargs):
        try:
            if not self.wfile.closed:
                self.wfile.flush()
                self.wfile.close()
        except Exception:
            pass
        self.rfile.close()


class InvalidRangeHeader(Exception):
    pass


def serve_file(manager):
    print("serving on http://localhost:%s" % manager.port)
    log.info("serving on http://localhost:%s" % manager.port)
    server_address = ('0.0.0.0', manager.port)
    httpd = ThreadedHTTPServer(server_address, VideoHandler)
    manager.httpd = httpd
    VideoHandler.manager = manager
    httpd.serve_forever()


def parse_range_header(range_header, total_length):
    """
    Return a 2-element tuple containing the requested range offsets
    in bytes.
    - range_header is the HTTP header sans the "Range:" prefix
    - total_length is the length in bytes of the requested resource
      (needed to calculate offsets for a 'n bytes from the end' request
    If no Range explicitly requested, returns (None, None)
    If Range header could not be parsed, raises InvalidRangeHeader
    (which could either be handled as a user
    request failure, or the same as if (None, None) was returned
    """
    if range_header is None or range_header == "":
        return (None, None)
    if not range_header.startswith("bytes="):
        log.error("Don't know how to parse Range: %s [1]" %
                  (range_header))
        raise InvalidRangeHeader("Don't know how to parse non-bytes Range: %s" %
                                 (range_header))
    regex = re.compile(r"^bytes=(\d*)\-(\d*)$")
    rangething = regex.search(range_header)
    if rangething:
        r1 = rangething.group(1)
        r2 = rangething.group(2)
        log.debug("Requested range is [%s]-[%s]" % (r1, r2))

        if r1 == "" and r2 == "":
            log.warning("Requested range is meaningless")
            raise InvalidRangeHeader("Requested range is meaningless")

        if r1 == "":
            # x bytes from the end of the file
            try:
                final_bytes = int(r2)
            except ValueError:
                raise InvalidRangeHeader("Invalid trailing range")
            return (total_length-final_bytes, total_length - 1)

        try:
            from_val = int(r1)
        except ValueError:
            raise InvalidRangeHeader("Invalid starting range value")
        if r2 != "":
            try:
                end_val = int(r2)
            except ValueError:
                raise InvalidRangeHeader("Invalid ending range value")
            return (from_val, end_val)
        else:
            return (from_val, total_length - 1)
    else:
        raise InvalidRangeHeader("Don't know how to parse Range: %s" %
                                 (range_header))



from KickassAPI import Search
import logging

from blessings import Terminal
from os import _exit
from time import time
from torrentmediasearcher import TorrentMediaSearcher

from touchandgo.decorators import with_config_dir
from touchandgo.download import DownloadManager
from touchandgo.helpers import get_settings
from touchandgo.history import History
from touchandgo.search.strike import StrikeAPI


log = logging.getLogger('touchandgo.main')
term = Terminal()


class SearchAndStream(object):
    """Searchs a torrent and start streaming it"""
    @with_config_dir
    def __init__(self, name, season=None, episode=None, sub_lang=None,
                 serve=False, quality=None, port=None, player=None,
                 search=None, use_cache=True):
        self.name = name
        self.season = season
        self.episode = episode
        self.sub_lang = sub_lang
        self.serve = serve
        self.quality = quality
        self.port = port
        self.player = player
        if search is None:
            settings = get_settings()
            search = settings.default_search_engine
        self.search_engine = search
        self.use_cache = use_cache

    @with_config_dir
    def download(self, results):
        print(term.bold('Downloading Magnet'))
        log.info("Processing magnet link")
        magnet = results['magnet']
        log.info("Magnet: %s", magnet)

        history = History(date=int(time()), name=self.name, season=self.season,
                          episode=self.episode, magnet=magnet)
        history.save()
        history.update()

        manager = DownloadManager(magnet, port=self.port, serve=self.serve,
                                  sub_lang=self.sub_lang, player=self.player)
        manager.start()

    def watch(self):
        print(term.bold('Welcome to Touchandgo'))
        try:
            if self.name.startswith('magnet'):
                results = {'magnet': self.name}
                self.download(results)
            else:
                history = None
                if self.use_cache:
                    history = History.one(name=self.name, season=self.season,
                                          episode=self.episode)
                if history is None or not hasattr(history, "magnet"):
                    self.search_magnet()
                else:
                    results = {'magnet': history.magnet}
                    self.download(results)

        except KeyboardInterrupt:
            log.info("Thanks for using Touchandgo")
            _exit(0)

    def search_magnet(self):
        print ("Searching magnet")
        log.info("Searching magnet")
        if self.search_engine == "kat":
            self.kat_search()
        elif self.search_engine == "strike":
            self.strike_search()
        else:
            self.tms_search()

    def get_search_string(self):
        search_string = self.name
        if self.season is not None and self.episode is not None:
            search_string += " s%se%s" % (str(self.season).zfill(2),
                                          str(self.episode).zfill(2))
        return search_string

    def kat_search(self):
        search_string = self.get_search_string()
        print("Searching '%s' on Kickass" % search_string)
        log.info("Searching %s on Kickass", search_string)
        results = Search(search_string).list()
        print(term.clear())
        print(term.bold("Touchandgo\n"))
        print(term.red("Kickass Torrents Results"))
        for i, result in enumerate(results, 1):
            option = term.cyan("%s) " % i)
            option += result.name + " "
            option += term.yellow(result.size)
            option += term.green(" S:" + result.seed)
            option += term.red(" L:" + result.leech)
            print(option)
        input_text = "Select which torrent you want to download (1-%d): " % \
            len(results)
        user_inuput = raw_input(input_text)
        try:
            opt = int(user_inuput) - 1
            if opt > len(results) or opt < 1:
                opt = 0
        except ValueError:
            opt = 0

        extra_tracker = "&tr=http%3A%2F%2Ftracker.nwps.ws%3A6969%2Fannounce"
        results = {'magnet': results[opt].magnet_link + extra_tracker}
        self.download(results)

    def strike_search(self):
        search_string = self.get_search_string()
        log.info("Searching %s on Strike", search_string)
        strike = StrikeAPI(search_string)
        magnet = strike.get_first_torrent_magnet()
        results = {'magnet': magnet}
        self.download(results)

    def tms_search(self):
        search = TorrentMediaSearcher
        if self.season is None and self.episode is None:
            search.request_movie_magnet('torrentproject', self.name,
                                        callback=self.download,
                                        quality=self.quality)
        else:
            if self.quality is None:
                quality = 'normal'
            else:
                quality = self.quality
            search.request_tv_magnet(provider='eztv', show=self.name,
                                     season=int(self.season),
                                     episode=int(self.episode),
                                     quality=quality, callback=self.download)



#! /usr/bin/env python2

# Get torrents from http://getstrike.net

import requests


class StrikeAPI():
    "Search torrents using http://getstrike.net/api/"
    def __init__(self, search_string):
        self.url = "http://getstrike.net/api/torrents"
        self.search_string = search_string
        self.headers = {'User-Agent': 'TouchAndGo (+https://github.com/touchandgo-devs/touchandgo)'}

    def search(self):
        search_url = '%s/search/' % (self.url)
        payload = {'q': self.search_string}
        response = requests.get(search_url, params=payload,
                                headers=self.headers)
        torrents = response.json()
        return torrents

    def get_first_torrent_magnet(self):
        torrents = self.search()
        first_torrent_hash = torrents[1][0]['torrent_hash']
        info_url = '%s/info/' % (self.url)
        payload = {'hashes': first_torrent_hash}
        response = requests.get(info_url, params=payload, headers=self.headers)
        magnet = response.json()[1][0]['magnet_uri']
        return magnet











#! /usr/bin/env python2
import argparse
import logging
import requests
import signal

from os import kill
from os.path import abspath, dirname, join

from subprocess import PIPE, STDOUT, Popen
from time import sleep
from flask import Flask, redirect, render_template, Response, request
from jinja2 import FileSystemLoader

from touchandgo.helpers import get_interface, LOCKFILE, get_free_port, \
    is_process_running
from touchandgo.history import History
from touchandgo.decorators import with_config_dir
from touchandgo.logger import log_set_up
from touchandgo.settings import DEBUG
from touchandgo.lock import Lock


log = logging.getLogger('touchandgo.proxy')


@with_config_dir
def serve(py_exec=None):
    log_set_up(DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument("--python", default="python2")
    args = parser.parse_args()

    py_exec = args.python

    app = Flask("touchandgo")
    template_path = join(dirname(__file__), "templates")
    app.jinja_loader = FileSystemLoader(template_path)

    def get_torrent(name, season=None, episode=None):
        interface = get_interface()
        path = abspath(__file__)
        dir_path = dirname(path)
        exec_ = join(dir_path, "__init__.py")
        command = '%s %s \"%s\" ' % (py_exec, exec_, name)
        if season is not None:
            command += "%s %s " % (season, episode)
        lock = Lock(LOCKFILE)
        if lock.is_same_file(name, season, episode) and \
                is_process_running(lock.get_pid()):
            data = lock._get_file_data()
            port = data[4]
        else:
            port = get_free_port()
            command += "--daemon --port %s " % port
            log.debug(command)
            process = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)
            sleep(1)

        redirect_url = "http://%s:%s" % (interface, port)
        serving = False
        while not serving:
            try:
                req = requests.get("%s/status" % redirect_url)
                serving = True
            except requests.ConnectionError, e:
                sleep(1)
        log.info("redirecting to %s" % redirect_url)
        return redirect(redirect_url)

    @app.route("/favicon.ico")
    def favicon():
        return ""

    @app.route("/magnet")
    def magnet():
        # Maybe someone wanted to search the movie "magnet", who knows
        return get_torrent(request.args.get('m', 'magnet'))

    @app.route("/<name>")
    @app.route("/<name>/<season>/<episode>")
    def redirect_to(name, season=None, episode=None):
        log.info("Requesting %s %s %s", name, season, episode)
        return get_torrent(name, season, episode)

    @app.route("/l/<name>/<season>/<episode>")
    @app.route("/l/<name>/<season>/<episode>/")
    def list_(name, season, episode):
        url = "http://%s:5000/%s/%s" % (get_interface(), name, season)
        episode = int(episode)

        template = app.jinja_env.get_template("m3u.j2")
        ret = template.render(episodes=range(episode, episode + 10),
                              series=name, season=episode, url=url)
        return Response(response=ret, status=200,
                        mimetype="application/x-mpegurl",
                        content_type="application/x-mpegurl")

    @app.route("/kill")
    def kill_():
        lock = Lock(LOCKFILE)
        try:
            kill(lock.get_pid(), signal.SIGQUIT)
        except Exception, e:
            pass
        return "OK"

    @app.route("/")
    def main():
        series = []
        keys = []
        items = History.many(sorted="-date,-season,-episode")

        # Filter duplicate entries in the history list
        for item in items:
            key = (item.name, item.season is not None)
            if key not in keys:
                series.append(item)
                keys.append(key)

        return render_template("main.html", items=series, magnet=request.args.get('m', ''))

    app.debug = True
    kill_()
    app.run(host="0.0.0.0")


if __name__ == '__main__':
    serve()



import pychromecast

from blessings import Terminal
from os import system
from time import sleep

from touchandgo.helpers import get_interface


class Output(object):
    def __init__(self, stream_url, sub_file, parent):
        self.stream_url = stream_url
        self.sub_file = sub_file
        self.parent = parent

    def run(self):
        player_command = self._player()
        if self.sub_file is not None:
            subs = self._subs()
        else:
            subs = ""

        command = "%s %s" % (player_command, subs)
        system(command)

    def _player(self):
        return ""

    def _subs(self):
        return ""


class VLCOutput(Output):
    def _player(self):
        return "vlc %s -q" % self.stream_url

    def _subs(self):
        subs = "--sub-file %s" % self.sub_file
        return subs


class OMXOutput(Output):
    def _player(self):
        return "omxplayer --live --timeout 360 -p -o hdmi %s" \
            % self.stream_url

    def _subs(self):
        subs = "--subtitle %s" % self.sub_file
        return subs


class CastOutput(Output):
    @classmethod
    def select_chromecast(cls):
        name = None
        chromecasts = pychromecast.get_chromecasts_as_dict().keys()

        if not chromecasts:
            return

        if len(chromecasts) > 1:
            term = Terminal()

            print(term.clear())
            print(term.bold("Touchandgo\n"))
            print(term.red("Chromecasts Availables"))
            for i, cc_name in enumerate(chromecasts, 1):
                option = term.cyan("%s) " % i)
                option += cc_name
                print(option)

            input_text = "Select which chromecast you want to cast (1-%d): " % \
                len(chromecasts)
            user_input = raw_input(input_text)

            try:
                opt = int(user_input) - 1
                if opt > len(chromecasts) or opt < 1:
                    opt = 0
            except ValueError:
                opt = 0

            name = chromecasts[opt]
        elif len(chromecasts) == 1:
            name = chromecasts[0]

        return name

    def run(self):
        chromecast = self.parent.chromecast
        device = pychromecast.get_chromecast(friendly_name=chromecast)
        interface = get_interface()
        guess = self.parent.guess(self.parent.get_video_path())
        device.play_media("http://%s:%s" % (interface, self.parent.port),
                          guess['mimetype'])
        while True:
            sleep(1)

    def __del__(self):
        self.chromecast.media_controller.stop()



import os
import logging
import signal
import socket

from daemon import DaemonContext
from datetime import datetime
from os import mkdir
from os.path import getmtime, exists, dirname, join
from shutil import copyfile

from altasetting import Settings
from netifaces import interfaces, ifaddresses
from ojota import set_data_source
from touchandgo.lock import Lock


LOCKFILE = "/tmp/touchandgo"
log = logging.getLogger('touchandgo.helpers')


def get_settings():
    settings_file = "%s/.touchandgo/settings.yaml" % os.getenv("HOME")
    default = join(dirname(__file__), "templates", "settings.yaml")

    set_config_dir()
    if not exists(settings_file):
        copyfile(default, settings_file)

    settings = Settings(settings_file, default)
    return settings


def get_free_port():
    socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_.bind(('localhost', 0))
    addr, port = socket_.getsockname()
    socket_.close()
    return port


def is_port_free(port):
    free = True
    socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        socket_.bind(('localhost', port))
    except socket.error:
        free = False
    socket_.close()

    return free


def get_interface():
    for ifaceName in interfaces():
        addresses = ifaddresses(ifaceName)
        for address in addresses.values():
            for item in address:
                if item.get('netmask') is not None and \
                        not item['addr'].startswith("127") and \
                        not item['addr'].startswith(":") and \
                        len(item['addr']) < 17:
                    return item['addr']


def is_process_running(process_id):
    try:
        os.kill(process_id, 0)
        return True
    except OSError:
        return False


def daemonize(args, callback):
    with DaemonContext():
        from touchandgo.logger import log_set_up
        log_set_up(True)
        log = logging.getLogger('touchandgo.daemon')
        log.info("running daemon")
        create_process = False
        lock = Lock(LOCKFILE, os.getpid(), args.name, args.sea_ep[0],
                    args.sea_ep[1], args.port)
        if lock.is_locked():
            log.debug("lock active")
            lock_pid = lock.get_pid()
            if not lock.is_same_file(args.name, args.sea_ep[0],
                                     args.sea_ep[1]) \
                    or not is_process_running(lock_pid):
                try:
                    log.debug("killing process %s" % lock_pid)
                    os.kill(lock_pid, signal.SIGQUIT)
                except OSError:
                    pass
                except TypeError:
                    pass
                lock.break_lock()
                create_process = True
        else:
            create_process = True

        if create_process:
            log.debug("creating proccess")
            lock.acquire()
            callback()
            lock.release()
        else:
            log.debug("same daemon process")


def get_lock_diff():
    timediff = 0
    try:
        now = datetime.now()
        timediff = now - datetime.fromtimestamp(getmtime(LOCKFILE + ".lock"))
        timediff = timediff.total_seconds()
    except OSError:
        pass
    return timediff


def set_config_dir():
    data_folder = "%s/.touchandgo" % os.getenv("HOME")
    if not exists(data_folder):
        mkdir(data_folder)

    set_data_source(data_folder)



import logging

from os.path import join

from babelfish import Language
from guessit import guess_video_info
from subliminal import download_best_subtitles
from subliminal.subtitle import get_subtitle_path
from subliminal.video import Video

from touchandgo.helpers import get_settings


log = logging.getLogger('touchandgo.download')


class SubtitleDownloader(object):
    def __init__(self, sub_lang):
        self.lang = sub_lang

    def download(self, video_file):
        subtitle = None
        settings = get_settings()
        download_dir = settings.save_path
        log.info("Downloading subtitle")
        filepath = join(download_dir, video_file[0])
        guess = guess_video_info(filepath, info=['filename'])
        video = Video.fromguess(filepath, guess)
        video.size = video_file[1]
        try:
            subtitle = download_best_subtitles([video], {Language(self.lang)},
                                               single=True)
        except ValueError:
            pass
        if subtitle is not None and len(subtitle):
            log.info("CLAH %s %s", download_dir, video.name)
            subtitle = get_subtitle_path(join(download_dir,
                                              video.name.replace("(", "\(")
                                              .replace(")", "\)")
                                              .replace(" ", "\ ")))
        log.info("video_file: %s, filepath: %s, guess: %s, video: %s, "
                 "subtitle: %s", video_file, filepath, guess, video, subtitle)
        return subtitle




import logging

from qtfaststart.processor import get_index
from qtfaststart.exceptions import FastStartException

from touchandgo.settings import SKIP_MOOV


log = logging.getLogger('touchandgo.helpers')


def have_moov(video_file):
    if not SKIP_MOOV:
        atoms = {}
        log.info("Checking moov data of %s", video_file)
        try:
            atom_data = get_index(open(video_file, 'rb'))
            for atom in atom_data:
                atoms[atom.name] = atom.position
            log.debug("moov:%(moov)s mdat:%(mdat)s ftyp:%(ftyp)s free:%(free)s",
                    atoms)

            if atoms['moov'] > atoms['mdat']:
                log.info("moov atom after mdat")
                return True, "after_mdat"
            else:
                log.info("moov atom before mdat")
                return True
        except FastStartException as e:
            log.error("Couldn't get Atoms data: %s", str(e))
            return False
    else:
        return True



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='Touchandgo',
    version='0.12.5',
    author='Felipe Lerena, Nicolás Demarchi',
    description='A python app and library to watch series magically',
    author_email='felipelerena@gmail.com - mail@gilgamezh.me',
    packages=['touchandgo'],
    scripts=[],
    url='https://github.com/touchandgo-devs/touchandgo/',
    license='LICENSE.txt',
    long_description=open('README.rst').read(),
    install_requires=['TorrentMediaSearcher==1.0.3',
                      'subliminal==0.7.4',
                      'netifaces==0.10.4',
                      'flask==0.10.1',
                      'simplejson==3.6.2',
                      'python-daemon==1.6.1',
                      'ojota==2.0.1',
                      'pyaml',
                      'colorama==0.3.2',
                      'requests==2.3.0',
                      'qtfaststart==1.8',
                      'KickassAPI',
                      'pyQuery',
                      'altasetting',
                      'blessings',
                      'guessit==0.6.2',
                      'pbr',
                      'PyChromecast==0.6.13',
                      ],
    entry_points={
        'console_scripts': ['touchandgo = touchandgo.main:main',
                            'tsproxy = touchandgo.tsproxy.__init__:serve']
    },
    package_data={
        'touchandgo': ['templates/*',
                       'download/*.py',
                       'search/*.py',
                       'tsproxy/*.py'],
    },
)



import logging
import sys

from logging.handlers import RotatingFileHandler
from os import mkdir
from os.path import exists

from touchandgo.helpers import get_settings
from touchandgo.settings import DEBUG


def log_set_up(verbose=False):
    settings = get_settings()
    if not exists(settings.save_path):
        mkdir(settings.save_path)

    logfile = "%s/touchandgo.log" % (settings.save_path)
    handler = RotatingFileHandler(logfile, maxBytes=1e6, backupCount=10)
    formatter = logging.Formatter("%(asctime)s  %(name)-22s  "
                                  "%(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.addHandler(handler)
    if DEBUG or verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    add_stdout_handler(logger, formatter)


def add_stdout_handler(logger, formatter):
    try:
        from rainbow_logging_handler import RainbowLoggingHandler
        handler = RainbowLoggingHandler(sys.stderr, color_funcName=(
            'black', 'black', True))
    except ImportError:
        handler = logging.StreamHandler()
        pass

    handler.setFormatter(formatter)
    logger.addHandler(handler)



import logging

from touchandgo.download.moov import have_moov
from touchandgo.helpers import get_settings


log = logging.getLogger('touchandgo.strategy')


class DefaultStrategy(object):
    def __init__(self, manager):
        settings = get_settings()
        self.settings = settings.strategy
        self.manager = manager
        self.piece_st = self.settings.piece_st
        self.last_piece_st = self.settings.last_piece_st
        self.holding_stream = True
        self.handle = manager.handle
        self.chunks_strat = self.settings.chunks_strat
        self.moov_downloaded = False
        self.download_lasts = False

    def initial(self):
        self.handle.set_sequential_download(True)
        status = self.handle.status()

        for i in range(self.piece_st):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 3000)

        self.chunks_strat = len(status.pieces) / 30

    def block_requested(self, block_requested):
        #log.debug("block requested %s" % block_requested)
        if not self.holding_stream:
            self.move_strategy(block_requested)

    def master(self):
        status = self.handle.status()
        pieces = status.pieces
        last_piece = len(status.pieces)
        first_n = pieces[:self.piece_st]
        if self.download_lasts:
            last_n = pieces[-self.last_piece_st:]
        else:
            last_n = []
        if all(first_n) and all(last_n):
            log.debug("first_n downloaded")
            if not self.moov_downloaded:
                video_file = self.manager.get_video_path()
                self.moov_downloaded = have_moov(video_file)
                log.debug("Moov Downloaded = %s", self.moov_downloaded)
            if self.moov_downloaded:
                if not self.settings.always_sequential:
                    self.handle.set_sequential_download(False)
                if not self.holding_stream:
                    self.piece_st += self.chunks_strat
                else:
                    self.holding_stream = False
                    self.manager.stream()
                if self.piece_st > 4:
                    self.piece_st += self.chunks_strat
                self.move_strategy(self.piece_st)
            else:
                log.debug("video doesn't have moov yet. Extending first_n")
                for i in range(last_piece - self.piece_st, last_piece):
                    self.handle.piece_priority(i, 7)
                    self.handle.set_piece_deadline(i, 3000)

                for i in range(self.piece_st):
                    self.handle.piece_priority(i, 3)
                    self.handle.set_piece_deadline(i, 3000)

                self.piece_st *= 2
                self.last_piece_st *= 2
                log.debug("piece_st = %s last_piece_st %s",
                          self.piece_st, self.last_piece_st)
                self.download_lasts = True

    def reset_priorities(self):
        for i in range(len(self.handle.status().pieces)):
            self.handle.piece_priority(i, 1)

    def move_strategy(self, first_block):
        log.debug("moving strategy %s" % first_block)
        self.reset_priorities()

        status = self.handle.status()
        last_piece = len(status.pieces) - 1

        last_chunks = first_block + self.chunks_strat
        if last_chunks > last_piece:
            last_chunks = last_piece
        for i in range(first_block, last_chunks):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 10000)

        last_chunks = first_block + self.chunks_strat * 2
        if last_chunks > last_piece:
            last_chunks = last_piece
        for i in range(first_block+self.chunks_strat, last_chunks):
            self.handle.piece_priority(i, 3)
            #self.handle.set_piece_deadline(i, 20000)

        self.piece_st = first_block + self.chunks_strat



#! /usr/bin/env python2
from __future__ import division

import argparse
import thread
import logging

from os import _exit
from os.path import join, exists
from time import sleep
from datetime import datetime, timedelta

from blessings import Terminal
from colorama import Fore
from guessit import guess_video_info
from libtorrent import add_magnet_uri, session, storage_mode_t

from touchandgo.constants import STATES
from touchandgo.download.strategy import DefaultStrategy
from touchandgo.download.subtitles import SubtitleDownloader
from touchandgo.helpers import is_port_free, get_free_port, get_settings
from touchandgo.logger import log_set_up
from touchandgo.output import VLCOutput, OMXOutput, CastOutput
from touchandgo.settings import DEBUG, WAIT_FOR_IT, DEFAULT_PORT
from touchandgo.stream_server import serve_file


log = logging.getLogger('touchandgo.download')
term = Terminal()


class DownloadManager(object):
    strategy_class = DefaultStrategy
    sub_downloader_class = SubtitleDownloader

    def __init__(self, magnet, port=None, sub_lang=None, serve=False,
                 player=None):
        self.settings = get_settings()
        self.magnet = magnet
        if port is None:
            port = DEFAULT_PORT
        port = int(port)
        if not is_port_free(port):
            port = get_free_port()

        log.info("[Magnet]: %s [Port]: %s [Sub_lang]: %s [Serve]: %s "
                 "[Player] %s ",
                 magnet, port, sub_lang, serve, player)

        self.port = port
        self.serve = serve
        if player is None:
            player = self.settings.players.default
        self.player = player

        # number of pieces to wait until start streaming
        # we are waiting untill all the first peices are downloaded
        # the biggest file which is supposed to be a video file
        self._video_file = None
        self.callback = serve_file
        self._served_blocks = None
        self.streaming = False

        self.stream_th = None
        self.player_th = None
        self.httpd = None
        self._guess = None

        self.init_handle()
        self.strategy = self.strategy_class(self)

        if sub_lang is not None:
            self.subtitle = self.sub_downloader_class(sub_lang)
        else:
            self.subtitle = None

    def init_handle(self):
        params = {
            "save_path": self.settings.save_path,
            "allocation": storage_mode_t.storage_mode_sparse,
            }
        self.session = session()
        self.handle = add_magnet_uri(self.session, str(self.magnet), params)

        up_limit = self.settings.limits.upload
        if up_limit:
            self.handle.set_upload_limit(up_limit)
        down_limit = self.settings.limits.download
        if down_limit:
            self.handle.set_download_limit(down_limit)

    @property
    def status(self):
        return self.handle.status()

    def start(self):
        try:
            self.start_time = datetime.now()
            self.session.listen_on(6881, 6891)
            self.session.start_dht()
            print("Downloading metadata")
            log.info("Downloading metadata")
            while not self.handle.has_metadata():
                print(term.clear())
                print(self.screen_data(DEBUG))
                elapsed_time = datetime.now() - self.start_time
                if elapsed_time > timedelta(minutes=2):
                    print("Torrent metadata not available")
                    return
                else:
                    sleep(.5)
            log.info("Starting download")
            self.strategy.initial()

            while True:
                is_seed = self.handle.is_seed()
                if not is_seed:
                    self.strategy.master()
                elif self.strategy.holding_stream:
                    self.strategy.holding_stream = False

                if not self.streaming and is_seed:
                    self.stream()

                print(term.clear())
                print(self.screen_data(DEBUG))
                sleep(1)
        except KeyboardInterrupt:
            if self.httpd is not None:
                self.httpd.shutdown()
            raise KeyboardInterrupt

    @property
    def video_file(self):
        if self._video_file is None:
            self._video_file = self.get_biggest_file()
            log.info("Video File: %s", self._video_file)
        return self._video_file

    def get_biggest_file(self):
        info = self.handle.get_torrent_info()
        biggest_file = ("", 0)
        files = info.files()
        for file_ in files:
            if file_.size > biggest_file[1]:
                biggest_file = [file_.path, file_.size]

        return biggest_file

    def wait_for_file(self):
        while not exists(join(self.settings.save_path, self.video_file[0])):
            sleep(WAIT_FOR_IT)

    def output(self, chromecast=None):
        self.chromecast = chromecast
        self.wait_for_file()
        stream_url = "http://localhost:%s" % self.port
        if self.subtitle is not None:
            subtitle = self.subtitle.download(self.video_file)
        else:
            subtitle = None

        players = {"vlc": VLCOutput,
                   "omxplayer": OMXOutput,
                   "chromecast": CastOutput
                   }
        output_class = players.get(self.player, VLCOutput)
        output = output_class(stream_url, subtitle, self)
        try:
            output.run()
        except KeyboardInterrupt:
            log.debug("Closing output thread")
        _exit(0)

    def stream(self):
        if self.callback is not None and not self.streaming:
            self.streaming = True
            pieces = self.status.pieces
            self._served_blocks = [False for i in range(len(pieces))]
            self.stream_th = thread.start_new_thread(self.callback, (self, ))
            if not self.serve:
                params = ()
                if self.player == "chromecast":
                    chromecast = CastOutput.select_chromecast()
                    params = (chromecast, )

                self.player_th = thread.start_new_thread(self.output, params)

    def block_served(self, block):
        self._served_blocks[block] = True

    def defrag(self):
        download_queue = self.handle.get_download_queue()
        downloading = [piece['piece_index'] for piece in download_queue]
        numerales = ""
        pieces = self.status.pieces
        for i, piece in enumerate(pieces):
            numeral = Fore.GREEN + "#" if piece else Fore.RED + "#"
            if i in downloading:
                numeral = Fore.YELLOW + "v"
            elif self._served_blocks is not None and self._served_blocks[i]:
                numeral = Fore.BLUE + ">"
            numeral += str(self.handle.piece_priority(i))
            numerales += numeral
        if numerales != "":
            numerales = term.magenta("\nPieces download state:\n" + numerales)
        return "%s\n" % numerales

    def screen_data(self, defrag=False):
        status = self.status
        text = Fore.WHITE
        text += term.bold('Touchandgo\n\n')
        text += STATES[status.state]
        text += term.yellow(' %.2f%% complete ' % (status.progress * 100))
        rates = self.rates()
        rate_text = '(down: %.1f kB/s up: %.1f kB/s peers: %d)\n' % \
                    (rates[0], rates[1], status.num_peers)
        text += term.green(rate_text)

        if True:#defrag:
            text += self.defrag()
            text += Fore.WHITE
            text += "\n\n"

        if self._video_file is not None:
            text += term.white("Serving ")
            text += term.yellow(self.video_file[0])
            text += term.white(" on ")
            text += term.blue("http://localhost:%s\n") % self.port

        text += term.white("Elapsed Time: ")
        text += term.red(self.elapsed_time())
        text += "\n"
        return text

    def elapsed_time(self):
        return str(datetime.now() - self.start_time).split(".")[0]

    def rates(self):
        return self.status.download_rate / 1000, self.status.upload_rate / 1000

    def get_video_path(self):
        video = self.video_file
        video_path = join(self.settings.save_path, video[0])
        return video_path

    def guess(self, path):
        if self._guess is None:
            self._guess = guess_video_info(path, info=['filename'])
        return self._guess


def main():
    log_set_up()
    parser = argparse.ArgumentParser()

    parser.add_argument("magnet", nargs='?', default=None)
    parser.add_argument("--sub", nargs='?', default=None)
    parser.add_argument("--serve", action="store_true")
    parser.add_argument("--port", "-p", default="8888")

    args = parser.parse_args()

    manager = DownloadManager(args.magnet, port=args.port, serve=args.serve,
                              sub_lang=args.sub)
    manager.start()

if __name__ == "__main__":
    main()



from ojota import Ojota


class History(Ojota):
    required_fields = ["date", "name", "season", "episode"]
    pk_field = "date"
    plural_name = "history"

    @property
    def next(self):
        if self.episode is not None:
            return int(self.episode) + 1
